#include <string.h>
#include "globals.h"
#include "serial.h"
#include "options.h"
#include "error_handlers.h"
#include "sensors.h"
#include "custom_gui.h"
#include "reset.h"
#include "version.h"
#include <stdio.h>

#define MSG_TOGGLE   MSG_USER
#define MSG_UPDATE   MSG_USER + 1
#define MSG_REFRESH  MSG_USER + 2

#define SENSORS_PER_PAGE      9
#define NUM_OF_RETRIES        2
#define SENSORS_TO_TIME_OUT   1 //number of sensors that need to time out before the warning will be issued
#define REFRESH_RATE_PRECISION 2 // how often the samples are taken, in milliseconds

// Sensor states:
#define SENSOR_OFF      0  // OFF,
#define SENSOR_ACTIVE   1  // ACTIVE (returned value),
#define SENSOR_NA       2  // NA (displays "N/A")
#define WINDOW_TITLE   "ScanTool.net " SCANTOOL_VERSION_EX_STR


typedef struct
{
   void (*formula)(int raw_data, char *buf);
   char label[32];
   char screen_buf[64];
   char pid[3];
   int enabled;
   int bytes; // number of data bytes expected from vehicle
   // dle added for PID support via PID 01 00, 01 20, and 01 40, so display can reflect if car supports PID
   int supported;
} SENSOR;

static void load_sensor_states();
static void save_sensor_states();
static void fill_sensors(int page_number);

static int reset_chip_proc(int msg, DIALOG *d, int c);
static int graph_proc(int msg, DIALOG *d, int c);
static int page_flipper_proc(int msg, DIALOG *d, int c);
static int sensor_proc(int msg, DIALOG *d, int c);
static int toggle_proc(int msg, DIALOG *d, int c);
static int toggle_all_proc(int msg, DIALOG *d, int c);
static int status_proc(int msg, DIALOG *d, int c);
static int page_number_proc(int msg, DIALOG *d, int c);
static int inst_refresh_rate_proc(int msg, DIALOG *d, int c);
static int avg_refresh_rate_proc(int msg, DIALOG *d, int c);
static int page_updn_handler_proc(int msg, DIALOG *d, int c);
static int genuine_proc(int msg, DIALOG *d, int c);

// Sensor formulae:
static void throttle_position_formula(int data, char *buf); // Page 1
static void engine_rpm_formula(int data, char *buf);
static void vehicle_speed_formula(int data, char *buf);
static void engine_load_formula(int data, char *buf);
static void timing_advance_formula(int data, char *buf);
static void intake_pressure_formula(int data, char *buf);
static void air_flow_rate_formula(int data, char *buf);
//static void fuel_system1_status_formula(int data, char *buf);
static void fuel_system2_status_formula(int data, char *buf);
static void short_term_fuel_trim_formula(int data, char *buf); // Page 2
static void long_term_fuel_trim_formula(int data, char *buf);
static void intake_air_temp_formula(int data, char *buf);
static void coolant_temp_formula(int data, char *buf);
static void fuel_pressure_formula(int data, char *buf);
static void secondary_air_status_formula(int data, char *buf);
static void pto_status_formula(int data, char *buf);
static void o2_sensor_formula(int data, char *buf);
void obd_requirements_formula(int data, char *buf);
// added 1/2/2003
static void engine_run_time_formula(int data, char *buf);
static void mil_distance_formula(int data, char *buf);
static void frp_relative_formula(int data, char *buf);
static void frp_widerange_formula(int data, char *buf);
static void o2_sensor_wrv_formula(int data, char *buf);
static void commanded_egr_formula(int data, char *buf);
static void egr_error_formula(int data, char *buf);
static void evap_pct_formula(int data, char *buf);
static void fuel_level_formula(int data, char *buf);
static void warm_ups_formula(int data, char *buf);
static void clr_distance_formula(int data, char *buf);
static void evap_vp_formula(int data, char *buf);
static void baro_pressure_formula(int data, char *buf);
static void o2_sensor_wrc_formula(int data, char *buf);
static void cat_temp_formula(int data, char *buf);
static void ecu_voltage_formula(int data, char *buf);
static void abs_load_formula(int data, char *buf);
static void eq_ratio_formula(int data, char *buf);
static void relative_tp_formula(int data, char *buf);
static void amb_air_temp_formula(int data, char *buf);
static void abs_tp_formula(int data, char *buf);
static void tac_pct_formula(int data, char *buf);
static void mil_time_formula(int data, char *buf);
static void clr_time_formula(int data, char *buf);
// dle added
static void oxygen_sensors_pres_formula(int data, char *buf);
static void monitor_status_DC_formula(int data, char *buf);



// variables
static int device_connected = FALSE;
static int reset_hardware = FALSE;
static int num_of_sensors = 0;
static int num_of_disabled_sensors = 0;
static int sensors_on_page = 0;
static int current_page = 0;
// dle added iGraphState to keep track of new Start/Stop/Graph button on Sensor screen
static int iGraphState = 0;
// dle added buf1 to use just for data (i.e. no text) for datalog.txt file
static char buf1[256];
static char title1[32];
static char title2[32];
static int iindex;

static float inst_refresh_rate = -1; // instantaneous refresh rate
static float avg_refresh_rate = -1;  // average refresh rate

static volatile int refresh_time; // time between sensor updates

static SENSOR sensorsbase[] =
{
   // formula                        // label            //screen_buffer      //pid  //enabled // bytes //supported
   // page 1
   { fuel_system2_status_formula,   "03: Fuel System 2 Status:",           "", "03",      0,    2,         0 },
   { engine_load_formula,           "04: Calculated Load Value:",          "", "04",      0,    1,         0 },
   { coolant_temp_formula,          "05: Coolant Temperature:",            "", "05",      0,    1,         0 },
   { short_term_fuel_trim_formula,  "06: S T Fuel Trim (Bank1):",          "", "06",      0,    2,         0 },
   { long_term_fuel_trim_formula,   "07: L T Fuel Trim (Bank1):",          "", "07",      0,    2,         0 },
   { short_term_fuel_trim_formula,  "08: S T Fuel Trim (Bank 2):",         "", "08",      0,    2,         0 },
   { long_term_fuel_trim_formula,   "09: L T Fuel Trim (Bank 2):",         "", "09",      0,    2,         0 },
   { fuel_pressure_formula,         "0A: Fuel Pressure (gauge):",          "", "0A",      0,    1,         0 },
   { intake_pressure_formula,       "0B: Intake Manifold Pressure:",       "", "0B",      0,    1,         0 },
   // Page 2
   { engine_rpm_formula,            "0C: Engine RPM:",                     "", "0C",      0,    2,         0 },
   { vehicle_speed_formula,         "0D: Vehicle Speed:",                  "", "0D",      0,    1,         0 },
   { timing_advance_formula,        "0E: Timing Advance (Cyl. #1):",       "", "0E",      0,    1,         0 },
   { intake_air_temp_formula,       "0F: Intake Air Temperature:",         "", "0F",      0,    1,         0 },
   { air_flow_rate_formula,         "10: Air Flow Rate (MAF sensor):",     "", "10",      0,    2,         0 },
   { throttle_position_formula,     "11: Absolute Throttle Position:",     "", "11",      0,    1,         0 },
   { secondary_air_status_formula,  "12: Secondary air status:",           "", "12",      0,    1,         0 },
   { oxygen_sensors_pres_formula,   "13: Oxygen Sensors Present:",         "", "13",      0,    1,         0 },
   { o2_sensor_formula,             "14: O2 Sensor 1, Bank 1:",            "", "14",      0,    2,         0 },
   // Page 3
   { o2_sensor_formula,             "15: O2 Sensor 2, Bank 1:",            "", "15",      0,    2,         0 },
   { o2_sensor_formula,             "16: O2 Sensor 3, Bank 1:",            "", "16",      0,    2,         0 },
   { o2_sensor_formula,             "17: O2 Sensor 4, Bank 1:",            "", "17",      0,    2,         0 },
   { o2_sensor_formula,             "18: O2 Sensor 1, Bank 2:",            "", "18",      0,    2,         0 },
   { o2_sensor_formula,             "19: O2 Sensor 2, Bank 2:",            "", "19",      0,    2,         0 },
   { o2_sensor_formula,             "1A: O2 Sensor 3, Bank 2:",            "", "1A",      0,    2,         0 },
   { o2_sensor_formula,             "1B: O2 Sensor 4, Bank 2:",            "", "1B",      0,    2,         0 },
   { obd_requirements_formula,      "1C: OBD conforms to:",                "", "1C",      0,    1,         0 },
   { oxygen_sensors_pres_formula,   "1D: Oxygen Sensors Present:",         "", "1D",      0,    1,         0 },
   // Page 4
   { pto_status_formula,            "1E: Power Take-Off Status:",          "", "1E",      0,    1,         0 },
   { engine_run_time_formula,       "1F: Time Since Engine Start:",        "", "1F",      0,    2,         0 },
   // note PID 20 is not used for display as it is PID for PID Support, PID 1F is SENSOR index (9*3) +2 -1 = 28
   { mil_distance_formula,          "21: Distance since MIL active:",      "", "21",      0,    2,         0 },
   { frp_relative_formula,          "22: FRP rel. to manifld vacuum:",     "", "22",      0,    2,         0 }, // fuel rail pressure relative to manifold vacuum
   { frp_widerange_formula,         "23: Fuel Pressure (gauge):",          "", "23",      0,    2,         0 }, // fuel rail pressure (gauge), wide range
   { o2_sensor_wrv_formula,         "24: O2 Sensor 1, Bank 1 (WR):",       "", "24",      0,    4,         0 }, // o2 sensors (wide range), voltage
   { o2_sensor_wrv_formula,         "25: O2 Sensor 2, Bank 1 (WR):",       "", "25",      0,    4,         0 },
   { o2_sensor_wrv_formula,         "26: O2 Sensor 3, Bank 1 (WR):",       "", "26",      0,    4,         0 },
   { o2_sensor_wrv_formula,         "27: O2 Sensor 4, Bank 1 (WR):",       "", "27",      0,    4,         0 },
   // Page 5
   { o2_sensor_wrv_formula,         "28: O2 Sensor 1, Bank 2 (WR):",       "", "28",      0,    4,         0 },
   { o2_sensor_wrv_formula,         "29: O2 Sensor 2, Bank 2 (WR):",       "", "29",      0,    4,         0 },
   { o2_sensor_wrv_formula,         "2A: O2 Sensor 3, Bank 2 (WR):",       "", "2A",      0,    4,         0 },
   { o2_sensor_wrv_formula,         "2B: O2 Sensor 4, Bank 2 (WR):",       "", "2B",      0,    4,         0 },
   { commanded_egr_formula,         "2C: Commanded EGR:",                  "", "2C",      0,    1,         0 },
   { egr_error_formula,             "2D: EGR Error:",                      "", "2D",      0,    1,         0 },
   { evap_pct_formula,              "2E: Commanded Evap Purge:",           "", "2E",      0,    1,         0 },
   { fuel_level_formula,            "2F: Fuel Level Input:",               "", "2F",      0,    1,         0 },
   { warm_ups_formula,              "30: Warm-ups since ECU reset:",       "", "30",      0,    1,         0 },
   // Page 6
   { clr_distance_formula,          "31: Distance since ECU reset:",       "", "31",      0,    2,         0 },
   { evap_vp_formula,               "32: Evap System Vapor Press:",        "", "32",      0,    2,         0 },
   { baro_pressure_formula,         "33: Barometric Pressure (abs):",      "", "33",      0,    1,         0 },
   { o2_sensor_wrc_formula,         "34: O2 Sensor 1, Bank 1 (WR):",       "", "34",      0,    4,         0 }, // o2 sensors (wide range), current
   { o2_sensor_wrc_formula,         "35: O2 Sensor 2, Bank 1 (WR):",       "", "35",      0,    4,         0 },
   { o2_sensor_wrc_formula,         "36: O2 Sensor 3, Bank 1 (WR):",       "", "36",      0,    4,         0 },
   { o2_sensor_wrc_formula,         "37: O2 Sensor 4, Bank 1 (WR):",       "", "37",      0,    4,         0 },
   { o2_sensor_wrc_formula,         "38: O2 Sensor 1, Bank 2 (WR):",       "", "38",      0,    4,         0 },
   { o2_sensor_wrc_formula,         "39: O2 Sensor 2, Bank 2 (WR):",       "", "39",      0,    4,         0 },
   // Page 7
   { o2_sensor_wrc_formula,         "3A: O2 Sensor 3, Bank 2 (WR):",       "", "3A",      0,    4,         0 },
   { o2_sensor_wrc_formula,         "3B: O2 Sensor 4, Bank 2 (WR):",       "", "3B",      0,    4,         0 },
   { cat_temp_formula,              "3C: CAT Temperature, B1S1:",          "", "3C",      0,    2,         0 },
   { cat_temp_formula,              "3D: CAT Temperature, B2S1:",          "", "3D",      0,    2,         0 },
   { cat_temp_formula,              "3E: CAT Temperature, B1S2:",          "", "3E",      0,    2,         0 },
   { cat_temp_formula,              "3F: CAT Temperature, B2S2:",          "", "3F",      0,    2,         0 },
   // note PID 40 is not used for display as it is PID for PID Support, PID 3F is SENSOR index (9*6) +6 -1 = 59
   { monitor_status_DC_formula,     "41: Monitor Status Drive Cycle:",     "", "41",      0,    2,         0 },
   { ecu_voltage_formula,           "42: ECU voltage:",                    "", "42",      0,    2,         0 },
   { abs_load_formula,              "43: Absolute Engine Load:",           "", "43",      0,    2,         0 },
   { eq_ratio_formula,              "44: Commanded Equiv Ratio:",          "", "44",      0,    2,         0 },
   // Page 8
   { relative_tp_formula,           "45: Relative Throttle Position:",     "", "45",      0,    1,         0 },
   { amb_air_temp_formula,          "46: Ambient Air Temperature:",        "", "46",      0,    1,         0 }, // same scaling as $0F
   { abs_tp_formula,                "47: Absolute Throttle Pos B:",        "", "47",      0,    1,         0 },
   { abs_tp_formula,                "48: Absolute Throttle Pos C:",        "", "48",      0,    1,         0 },
   { abs_tp_formula,                "49: Accelerator Pedal Pos D:",        "", "49",      0,    1,         0 },
   { abs_tp_formula,                "4A: Accelerator Pedal Pos E:",        "", "4A",      0,    1,         0 },
   { abs_tp_formula,                "4B: Accelerator Pedal Pos F:",        "", "4B",      0,    1,         0 },
   { tac_pct_formula,               "4C: Comm. Throttle Act Cntrl:",       "", "4C",      0,    1,         0 }, // commanded TAC
   { mil_time_formula,              "4D: Engine run while MIL on:",        "", "4D",      0,    2,         0 }, // minutes run by the engine while MIL activated
   // Page 9
   { clr_time_formula,              "4E: Time since DTCs cleared:",        "", "4E",      0,    2,         0 },
   { NULL,                          "",                                    "", "",        0,    0,         0 }
};

static SENSOR sensors[] =
{
   // formula                        // label            //screen_buffer      //pid  //enabled // bytes //supported
   // page 1
   { NULL,                          "",                                    "", "",        0,    0,         0 },
   { NULL,                          "",                                    "", "",        0,    0,         0 },
   { NULL,                          "",                                    "", "",        0,    0,         0 },
   { NULL,                          "",                                    "", "",        0,    0,         0 },
   { NULL,                          "",                                    "", "",        0,    0,         0 },
   { NULL,                          "",                                    "", "",        0,    0,         0 },
   { NULL,                          "",                                    "", "",        0,    0,         0 },
   { NULL,                          "",                                    "", "",        0,    0,         0 },
   { NULL,                          "",                                    "", "",        0,    0,         0 },
   // page 2
   { NULL,                          "",                                    "", "",        0,    0,         0 },
   { NULL,                          "",                                    "", "",        0,    0,         0 },
   { NULL,                          "",                                    "", "",        0,    0,         0 },
   { NULL,                          "",                                    "", "",        0,    0,         0 },
   { NULL,                          "",                                    "", "",        0,    0,         0 },
   { NULL,                          "",                                    "", "",        0,    0,         0 },
   { NULL,                          "",                                    "", "",        0,    0,         0 },
   { NULL,                          "",                                    "", "",        0,    0,         0 },
   { NULL,                          "",                                    "", "",        0,    0,         0 },
   // page 3
   { NULL,                          "",                                    "", "",        0,    0,         0 },
   { NULL,                          "",                                    "", "",        0,    0,         0 },
   { NULL,                          "",                                    "", "",        0,    0,         0 },
   { NULL,                          "",                                    "", "",        0,    0,         0 },
   { NULL,                          "",                                    "", "",        0,    0,         0 },
   { NULL,                          "",                                    "", "",        0,    0,         0 },
   { NULL,                          "",                                    "", "",        0,    0,         0 },
   { NULL,                          "",                                    "", "",        0,    0,         0 },
   { NULL,                          "",                                    "", "",        0,    0,         0 },
   // page 4
   { NULL,                          "",                                    "", "",        0,    0,         0 },
   { NULL,                          "",                                    "", "",        0,    0,         0 },
   { NULL,                          "",                                    "", "",        0,    0,         0 },
   { NULL,                          "",                                    "", "",        0,    0,         0 },
   { NULL,                          "",                                    "", "",        0,    0,         0 },
   { NULL,                          "",                                    "", "",        0,    0,         0 },
   { NULL,                          "",                                    "", "",        0,    0,         0 },
   { NULL,                          "",                                    "", "",        0,    0,         0 },
   { NULL,                          "",                                    "", "",        0,    0,         0 },
   // page 5
   { NULL,                          "",                                    "", "",        0,    0,         0 },
   { NULL,                          "",                                    "", "",        0,    0,         0 },
   { NULL,                          "",                                    "", "",        0,    0,         0 },
   { NULL,                          "",                                    "", "",        0,    0,         0 },
   { NULL,                          "",                                    "", "",        0,    0,         0 },
   { NULL,                          "",                                    "", "",        0,    0,         0 },
   { NULL,                          "",                                    "", "",        0,    0,         0 },
   { NULL,                          "",                                    "", "",        0,    0,         0 },
   { NULL,                          "",                                    "", "",        0,    0,         0 },
   // page 6
   { NULL,                          "",                                    "", "",        0,    0,         0 },
   { NULL,                          "",                                    "", "",        0,    0,         0 },
   { NULL,                          "",                                    "", "",        0,    0,         0 },
   { NULL,                          "",                                    "", "",        0,    0,         0 },
   { NULL,                          "",                                    "", "",        0,    0,         0 },
   { NULL,                          "",                                    "", "",        0,    0,         0 },
   { NULL,                          "",                                    "", "",        0,    0,         0 },
   { NULL,                          "",                                    "", "",        0,    0,         0 },
   { NULL,                          "",                                    "", "",        0,    0,         0 },
   // page 7
   { NULL,                          "",                                    "", "",        0,    0,         0 },
   { NULL,                          "",                                    "", "",        0,    0,         0 },
   { NULL,                          "",                                    "", "",        0,    0,         0 },
   { NULL,                          "",                                    "", "",        0,    0,         0 },
   { NULL,                          "",                                    "", "",        0,    0,         0 },
   { NULL,                          "",                                    "", "",        0,    0,         0 },
   { NULL,                          "",                                    "", "",        0,    0,         0 },
   { NULL,                          "",                                    "", "",        0,    0,         0 },
   { NULL,                          "",                                    "", "",        0,    0,         0 },
   // page 8
   { NULL,                          "",                                    "", "",        0,    0,         0 },
   { NULL,                          "",                                    "", "",        0,    0,         0 },
   { NULL,                          "",                                    "", "",        0,    0,         0 },
   { NULL,                          "",                                    "", "",        0,    0,         0 },
   { NULL,                          "",                                    "", "",        0,    0,         0 },
   { NULL,                          "",                                    "", "",        0,    0,         0 },
   { NULL,                          "",                                    "", "",        0,    0,         0 },
   { NULL,                          "",                                    "", "",        0,    0,         0 },
   { NULL,                          "",                                    "", "",        0,    0,         0 },
   // page 9
   { NULL,                          "",                                    "", "",        0,    0,         0 },
   { NULL,                          "",                                    "", "",        0,    0,         0 },
   { NULL,                          "",                                    "", "",        0,    0,         0 },
   { NULL,                          "",                                    "", "",        0,    0,         0 },
   { NULL,                          "",                                    "", "",        0,    0,         0 },
   { NULL,                          "",                                    "", "",        0,    0,         0 },
   { NULL,                          "",                                    "", "",        0,    0,         0 },
   { NULL,                          "",                                    "", "",        0,    0,         0 },
   { NULL,                          "",                                    "", "",        0,    0,         0 },
};



static DIALOG sensor_dialog[] =
{
   /* (proc)                (x)  (y)  (w)  (h) (fg)     (bg)          (key)(flags)  (d1) (d2)   (dp)                (dp2) (dp3) */
   { d_clear_proc,           0,   0,   0,   0,  0,       C_WHITE,       0,    0,      0,   0,   NULL,               NULL, NULL },
   { page_updn_handler_proc, 0,   0,   0,   0,  0,       0,             0,    0,      0,   0,   0,                  NULL, NULL },
   { d_shadow_box_proc,      40,  20,  560, 56, C_BLACK, C_LIGHT_GRAY,  0,    0,      0,   0,   NULL,               NULL, NULL },
   { d_rtext_proc,           50,  25,  96,  20, C_BLACK, C_TRANSP,      0,    0,      0,   0,   "Port Status:",     NULL, NULL },
   { status_proc,            172, 25,  288, 20, C_BLACK, C_LIGHT_GRAY,  0,    0,      0,   0,   NULL,               NULL, NULL },
   { reset_chip_proc,        464, 28,  126, 40, C_BLACK, C_GREEN,       'r',  D_EXIT, 0,   0,   "&Reset Interface", NULL, NULL },
   { d_rtext_proc,           50,  51,  96,  20, C_BLACK, C_TRANSP,      0,    0,      0,   0,   "Refresh rate:",    NULL, NULL },
   { inst_refresh_rate_proc, 154, 51,  174, 20, C_BLACK, C_LIGHT_GRAY,  0,    0,      0,   0,   NULL,               NULL, NULL },
   { avg_refresh_rate_proc,  328, 51,  132, 20, C_BLACK, C_LIGHT_GRAY,  0,    0,      0,   0,   NULL,               NULL, NULL },
   { d_box_proc,             40,  87,  560, 32, C_BLACK, C_LIGHT_GRAY,  0,    0,      0,   0,   NULL,               NULL, NULL },
   { toggle_proc,            45,  91,  45,  24, C_BLACK, C_WHITE,       0,    D_EXIT, 0,   0,   NULL,               NULL, NULL },
   { sensor_proc,            95,  94,  504, 24, C_BLACK, C_LIGHT_GRAY,  0,    0,      0,   0,   NULL,               NULL, NULL },
   { d_box_proc,             40,  123, 560, 32, C_BLACK, C_LIGHT_GRAY,  0,    0,      0,   0,   NULL,               NULL, NULL },
   { toggle_proc,            45,  127, 45,  24, C_BLACK, C_WHITE,       0,    D_EXIT, 1,   0,   NULL,               NULL, NULL },
   { sensor_proc,            95,  130, 504, 24, C_BLACK, C_LIGHT_GRAY,  0,    0,      1,   0,   NULL,               NULL, NULL },
   { d_box_proc,             40,  159, 560, 32, C_BLACK, C_LIGHT_GRAY,  0,    0,      0,   0,   NULL,               NULL, NULL },
   { toggle_proc,            45,  163, 45,  24, C_BLACK, C_WHITE,       0,    D_EXIT, 2,   0,   NULL,               NULL, NULL },
   { sensor_proc,            95,  166, 504, 24, C_BLACK, C_LIGHT_GRAY,  0,    0,      2,   0,   NULL,               NULL, NULL },
   { d_box_proc,             40,  195, 560, 32, C_BLACK, C_LIGHT_GRAY,  0,    0,      0,   0,   NULL,               NULL, NULL },
   { toggle_proc,            45,  199, 45,  24, C_BLACK, C_WHITE,       0,    D_EXIT, 3,   0,   NULL,               NULL, NULL },
   { sensor_proc,            95,  202, 504, 24, C_BLACK, C_LIGHT_GRAY,  0,    0,      3,   0,   NULL,               NULL, NULL },
   { d_box_proc,             40,  231, 560, 32, C_BLACK, C_LIGHT_GRAY,  0,    0,      0,   0,   NULL,               NULL, NULL },
   { toggle_proc,            45,  235, 45,  24, C_BLACK, C_WHITE,       0,    D_EXIT, 4,   0,   NULL,               NULL, NULL },
   { sensor_proc,            95,  238, 504, 24, C_BLACK, C_LIGHT_GRAY,  0,    0,      4,   0,   NULL,               NULL, NULL },
   { d_box_proc,             40,  267, 560, 32, C_BLACK, C_LIGHT_GRAY,  0,    0,      0,   0,   NULL,               NULL, NULL },
   { toggle_proc,            45,  271, 45,  24, C_BLACK, C_WHITE,       0,    D_EXIT, 5,   0,   NULL,               NULL, NULL },
   { sensor_proc,            95,  274, 504, 24, C_BLACK, C_LIGHT_GRAY,  0,    0,      5,   0,   NULL,               NULL, NULL },
   { d_box_proc,             40,  303, 560, 32, C_BLACK, C_LIGHT_GRAY,  0,    0,      0,   0,   NULL,               NULL, NULL },
   { toggle_proc,            45,  307, 45,  24, C_BLACK, C_WHITE,       0,    D_EXIT, 6,   0,   NULL,               NULL, NULL },
   { sensor_proc,            95,  310, 504, 24, C_BLACK, C_LIGHT_GRAY,  0,    0,      6,   0,   NULL,               NULL, NULL },
   { d_box_proc,             40,  339, 560, 32, C_BLACK, C_LIGHT_GRAY,  0,    0,      0,   0,   NULL,               NULL, NULL },
   { toggle_proc,            45,  343, 45,  24, C_BLACK, C_WHITE,       0,    D_EXIT, 7,   0,   NULL,               NULL, NULL },
   { sensor_proc,            95,  346, 504, 24, C_BLACK, C_LIGHT_GRAY,  0,    0,      7,   0,   NULL,               NULL, NULL },
   { d_box_proc,             40,  375, 560, 32, C_BLACK, C_LIGHT_GRAY,  0,    0,      0,   0,   NULL,               NULL, NULL },
   { toggle_proc,            45,  379, 45,  24, C_BLACK, C_WHITE,       0,    D_EXIT, 8,   0,   NULL,               NULL, NULL },
   { sensor_proc,            95,  382, 504, 24, C_BLACK, C_LIGHT_GRAY,  0,    0,      8,   0,   NULL,               NULL, NULL },
   { toggle_all_proc,        40,  420, 100, 40, C_BLACK, C_DARK_YELLOW, 'a',  D_EXIT, 0,   0,   "&All ON",          NULL, NULL },
   { graph_proc,             150, 420, 100, 40, C_BLACK, C_RED,         0,  D_EXIT,   0,   0,   "Start Data",       NULL, NULL },
   { d_shadow_box_proc,      260, 420, 230, 40, C_BLACK, C_LIGHT_GRAY,  0,    0,      0,   0,   NULL,               NULL, NULL },
   { d_button_proc,          500, 420, 100, 40, C_BLACK, C_DARK_YELLOW, 'm',  D_EXIT, 0,   0,   "&Main Menu",       NULL, NULL },
   { st_ctext_proc,          300, 422, 38,  20, C_BLACK, C_TRANSP,      0,    0,      0,   0,   "Page",             NULL, NULL },
   { page_flipper_proc,      340, 425, 75,  30, C_BLACK, C_DARK_YELLOW, 'p',  D_EXIT, -1,  0,   "&Previous",        NULL, NULL },
   { page_flipper_proc,      425, 425, 55,  30, C_BLACK, C_GREEN,       'x',  D_EXIT, 1,   0,   "Ne&xt",            NULL, NULL },
   { page_number_proc,       300, 440, 36,  18, C_BLACK, C_LIGHT_GRAY,  0,    0,      0,   0,   NULL,               NULL, NULL },
   { genuine_proc,           0,   0,   0,   0,  0,       0,             0,    0,      0,   0,   NULL,               NULL, NULL },
   { d_yield_proc,           0,   0,   0,   0,  0,       0,             0,    0,      0,   0,   NULL,               NULL, NULL },
   { NULL,                   0,   0,   0,   0,  0,       0,             0,    0,      0,   0,   NULL,               NULL, NULL }
};

void inc_refresh_time(void)
{
   refresh_time++;
}
END_OF_FUNCTION(inc_refresh_time);


int display_sensor_dialog(int reset)
{
   char PIDtextstring[128];
   int i, isensor;
   int ret;
   unsigned int iPIDHex1, iPIDHex2, iPIDHex3;
   unsigned int iPIDHex4 = 0;
   char buf[256];
   char cmd[8];
   char PIDByte1String[4] = "";
   char PIDByte2String[4] = "";
   char PIDByte3String[4] = "";
   char PIDByte4String[4] = "";
   int response_status = EMPTY; //status of the response: EMPTY, DATA, PROMPT
   char tempString[16];

   for (i = 0; sensorsbase[i].formula; i++);
   num_of_sensors = i;

   current_page = 0;
   if (reset)
      reset_hardware = TRUE;

   // dle adding code to show only support PIDs, doing this by brute force, i.e. just send command wait 1 second
   set_window_title("Initializing..");
   sprintf(cmd, "01 00");                // 00 is send back support PIDs for 00 - 20
   send_command(cmd);                    // send command
   sleep(1);                             // wait for response
   response_status = read_comport(buf);  // read comport
   if (response_status == PROMPT) {      // means we received entire message

      sprintf(PIDtextstring, "\nReading Supported PIDS via PID 00 ..... %s", buf);
      write_log(PIDtextstring);

      // now process the PID 00 response to populate supported PIDS structure, typical response is PID 00 = 41 00 BE 3D B8 10
      // note buf is string and need to convert to hex number first, need from position 6, 9, 12, 15; 2 chars each
      strncpy(PIDByte1String, (buf+6), 2);
      strncpy(PIDByte2String, (buf+9), 2);
      strncpy(PIDByte3String, (buf+12), 2);
      strncpy(PIDByte4String, (buf+15), 2);
      write_log("Converting to hex numbers.... ");
      write_log(PIDByte1String);
      write_log(PIDByte2String);
      write_log(PIDByte3String);
      write_log(PIDByte4String);

      // now convert the hex string to an int
      iPIDHex1 = strtol(PIDByte1String, NULL, 16);
      iPIDHex2 = strtol(PIDByte2String, NULL, 16);
      iPIDHex3 = strtol(PIDByte3String, NULL, 16);
      iPIDHex4 = strtol(PIDByte4String, NULL, 16);

      write_log("\nConverting to int number..... ");
      sprintf(tempString, "%x", iPIDHex1);
      write_log(tempString);
      sprintf(tempString, "%x", iPIDHex2);
      write_log(tempString);
      sprintf(tempString, "%x", iPIDHex3);
      write_log(tempString);
      sprintf(tempString, "%x", iPIDHex4);
      write_log(tempString);

      // process supported PIDs for PID 00 response, note SENSORS array starts at PID 3 for index 0
      for (i=8; i>2; i--) {
         sensorsbase[i-3].supported    = (iPIDHex1 >> (8 - i)) & 0x01;
      }
      for (i=8; i>0; i--) {
         sensorsbase[i+8-3].supported  = (iPIDHex2 >> (8 - i)) & 0x01;
         sensorsbase[i+16-3].supported = (iPIDHex3 >> (8 - i)) & 0x01;
      }
      // can't do last one as this is PID 20, and PID 20 is not part of SENSOR structure
      for (i=7; i>0; i--) {
         sensorsbase[i+24-3].supported = (iPIDHex4 >> (8 - i)) & 0x01;
      }
      // sanity check, last filled SENSOR value is PID 1F, which corresponds to index 31-3 = 28
   }
   else {
      sprintf(PIDtextstring, "\nNo Support for PID 00.....");
      write_log(PIDtextstring);
      reset_chip();
   }

   // PID 00 reflects whether PID 20 is supported, so determine PID 20 support before populating PID 20-3F support

   if (iPIDHex4 & 0x01) {

      set_window_title("Initializing....");
      sprintf(cmd, "01 20");                // 20 is send back support PIDs for 20 - 40
      send_command(cmd);                    // send command
      sleep(1);                             // wait for response
      response_status = read_comport(buf);  // read comport
      if (response_status == PROMPT) {      // means we received entire message

         sprintf(PIDtextstring, "\nReading Supported PIDS via PID 20 ..... %s", buf);
         write_log(PIDtextstring);
         // now process the PID 20 response to populate supported PIDS structure
         // note buf is string and need to convert to hex number first, need from position 6, 9, 12, 15; 2 chars each
         strncpy(PIDByte1String, (buf+6), 2);
         strncpy(PIDByte2String, (buf+9), 2);
         strncpy(PIDByte3String, (buf+12), 2);
         strncpy(PIDByte4String, (buf+15), 2);
         write_log("Converting to hex numbers.... ");
         write_log(PIDByte1String);
         write_log(PIDByte2String);
         write_log(PIDByte3String);
         write_log(PIDByte4String);

         // now convert the hex string to an int
         iPIDHex1 = strtol(PIDByte1String, NULL, 16);
         iPIDHex2 = strtol(PIDByte2String, NULL, 16);
         iPIDHex3 = strtol(PIDByte3String, NULL, 16);
         iPIDHex4 = strtol(PIDByte4String, NULL, 16);

         write_log("\nConverting to int number..... ");
         sprintf(tempString, "%x", iPIDHex1);
         write_log(tempString);
         sprintf(tempString, "%x", iPIDHex2);
         write_log(tempString);
         sprintf(tempString, "%x", iPIDHex3);
         write_log(tempString);
         sprintf(tempString, "%x", iPIDHex4);
         write_log(tempString);

         // process supported PIDs for PID 20 response, note SENSOR array skips PID 20, so need to skip 1 more, i.e 3+1 now
         for (i=8; i>1; i--) {
            sensorsbase[i+32-4].supported = (iPIDHex1 >> (8 - i)) & 0x01;
         }
         for (i=8; i>0; i--) {
            sensorsbase[i+40-4].supported = (iPIDHex2 >> (8 - i)) & 0x01;
            sensorsbase[i+48-4].supported = (iPIDHex3 >> (8 - i)) & 0x01;
         }
         // can't do last one as this is PID 20, and PID 20 is not part of SENSOR structure
         for (i=7; i>0; i--) {
            sensorsbase[i+56-4].supported = (iPIDHex4 >> (8 - i)) & 0x01;
         }
         // sanity check, last filled SENSOR value is PID 3F, which corresponds to index 56+7-4 = 59

      }
      else {
         sprintf(PIDtextstring, "\nNo Support for PID 20.....");
         write_log(PIDtextstring);
         reset_chip();
      }
   }

   // now populate sensors with only supported PID values
   isensor = 0;  // use as index to correct sensor structure
   for (i=0; i<num_of_sensors; i++){
       if ( sensorsbase[i].supported == 1){
           sensors[isensor] = sensorsbase[i];
           isensor++;
       }
   }

   // need to correct number of sensors now, as all won't be populated
   for (i = 0; sensors[i].formula; i++);
   num_of_sensors = i;


   // continue with rest of init
   load_sensor_states();
   fill_sensors(0);

   LOCK_VARIABLE(refresh_time);
   LOCK_FUNCTION(inc_refresh_time);

   set_window_title(WINDOW_TITLE);

   ret = do_dialog(sensor_dialog, -1);
   save_sensor_states();

   return ret;
}


static void calculate_refresh_rate(int sensor_state)
{
   static int initialization_occured = FALSE;
   static int num_of_sensors_off = 0;
   static int reset_on_all_off_occured = FALSE;
   static int sensors_on_counter = 0;
   static float avg_refresh_rate_accumulator = 0;

   if (!initialization_occured) // we received our first ">", initialize...
   {
      if (sensor_state == SENSOR_ACTIVE) // if we received HEX data
      {
         refresh_time = 0; // reset the time
         install_int(inc_refresh_time, REFRESH_RATE_PRECISION); // install handler
         initialization_occured = TRUE;
      }
   }
   else  // if this is not our first ">"
   {
      if ((num_of_sensors_off >= sensors_on_page) && !reset_on_all_off_occured) // if all sensors on page are OFF
      {
        inst_refresh_rate = -1;
         avg_refresh_rate = -1;
         reset_on_all_off_occured = TRUE;
         broadcast_dialog_message(MSG_REFRESH, 0);
      }
      else  // if num_of_sensors_off < sensors_on_page
      {
         if (sensor_state != SENSOR_OFF)
         {
            reset_on_all_off_occured = FALSE;
            inst_refresh_rate = 1/(refresh_time*REFRESH_RATE_PRECISION*0.001);

            if (sensors_on_counter < (sensors_on_page - num_of_disabled_sensors))
            {
               sensors_on_counter++;
               avg_refresh_rate_accumulator += inst_refresh_rate;
            }
            else
            {
               avg_refresh_rate = avg_refresh_rate_accumulator/sensors_on_counter;
               avg_refresh_rate_accumulator = 0;
               sensors_on_counter = 0;
            }

            if (sensor_state == SENSOR_ACTIVE) // if we got response from ECU
               refresh_time = 0; // reset time
            broadcast_dialog_message(MSG_REFRESH, 0);
         }
      }
   }
}


int reset_chip_proc(int msg, DIALOG *d, int c)
{
   int ret = d_button_proc(msg, d, c);

   if (ret == D_CLOSE)
   {
      if (comport.status == READY)
         reset_hardware = TRUE;
      else
         alert("Port is not ready.", NULL, NULL, "OK", NULL, 0, 0);
      ret = D_REDRAWME;
   }

   return ret;
}

// dle changed options_proc for new graph_proc, that is used old Options button to add new graphing capabilties
// note graph_proc is i=37 for sensor_dialog structure, need to change sensor_dialog[i].dp for text, .bg for background color
// need int iGraphState value to keep track of 3 states: 0=Start, 1=Stop, 2=Graph
int graph_proc(int msg, DIALOG *d, int c)
{
   int ret = d_button_proc(msg, d, c);
   char ytitle[128], plotcommand[256];

   if (ret == D_CLOSE)
   {
      // transitions to next state occur only when GUI button is pressed
      switch (iGraphState) {
         // current case 0 - previously normal operation, no stored data, just GUI data available
         // current button says 'Start Data', next state is 'Stop Data'
         case 0 :
            // need to remove previous datalogfile data
            remove(datalog_file_name);
            iGraphState = 1;
            sensor_dialog[37].dp = "Stop Data";
            sensor_dialog[37].bg = C_DARK_YELLOW;
            break;

         // current case 1 - store collected data in datafile for use in graphing later
         // current button says 'Stop Data', next state is 'Graph Data'
         case 1 :
            iGraphState = 2;
            sensor_dialog[37].bg = C_GREEN;
            sensor_dialog[37].dp = "Graph Data";
            break;

         // current case 2 - stop data collecting
         // current button says 'Graph Data', next state is graph the stored datafile, then 'Start Data'
         case 2 :
            iGraphState = 0;
            sensor_dialog[37].bg = C_RED;
            sensor_dialog[37].dp = "Start Data";
            // add graphing code here...persistent keeps graph open
            FILE *pipe_gp = popen("gnuplot -persistent", "w");
            fputs("set xlabel 'Seconds' \n", pipe_gp);
            // set label based on sensor index     // dle
            sprintf(ytitle, "set ylabel '%s' \n", sensors[iindex].label);
            fputs(ytitle, pipe_gp);
            // gnuplot ignores second graph if no data, so we can plot both all the tim e
            sprintf(plotcommand, "plot 'datalog.txt' using 1:2 smooth csplines title '%s', \
                            'datalog.txt' using 1:3 smooth csplines title '%s' \n", title1, title2);
            fputs(plotcommand, pipe_gp);
            pclose(pipe_gp);
            break;
       }
   }

   ret = D_REDRAWME;
   return ret;
}


int page_updn_handler_proc(int msg, DIALOG *d, int c)
{
   if ((msg == MSG_XCHAR) && ((c>>8) == KEY_PGUP || (c>>8) == KEY_PGDN))
   {
      if ((c>>8) == KEY_PGUP)
         simulate_keypress('p');
      if ((c>>8) == KEY_PGDN)
         simulate_keypress('x');

      return D_USED_CHAR;
   }

   return D_O_K;
}

int inst_refresh_rate_proc(int msg, DIALOG *d, int c)
{
   switch (msg)
   {
      case MSG_START:
         if ((d->dp = calloc(64, sizeof(char))) == NULL)
            fatal_error("Could not allocate enough memory for instantaneous refresh rate control");
         // Fall through

      case MSG_REFRESH:
      case MSG_UPDATE:
         if (inst_refresh_rate >= 0)
           sprintf(d->dp, "Instantaneous: %.2fHz", inst_refresh_rate);
         else
            sprintf(d->dp, "Instantaneous: N/A");
         d->flags |= D_DIRTY;
         break;

      case MSG_DRAW:
         rectfill(screen, d->x, d->y, d->x+d->w-1, d->y+d->h-1, d->bg);  // clear the element
         break;

      case MSG_END:
         free(d->dp);
         d->dp = NULL;
         break;
   }

   return d_text_proc(msg, d, c);
}


int avg_refresh_rate_proc(int msg, DIALOG *d, int c)
{
    switch (msg)
   {
      case MSG_START:
         if ((d->dp = calloc(64, sizeof(char))) == NULL)
            fatal_error("Could not allocate enough memory for average refresh rate control");
         // Fall through

      case MSG_REFRESH:
      case MSG_UPDATE:
         if (avg_refresh_rate >= 0)
           sprintf(d->dp, "Average: %.2fHz", avg_refresh_rate);
         else
            sprintf(d->dp, "Average: N/A");
         d->flags |= D_DIRTY;
         break;

      case MSG_DRAW:
         rectfill(screen, d->x, d->y, d->x+d->w-1, d->y+d->h-1, d->bg);  // clear the element
         break;

      case MSG_END:
         free(d->dp);
         d->dp = NULL;
         break;
   }

   return d_text_proc(msg, d, c);
}

/* ---- DO NOT TRANSLATE FROM HERE ---- */
void load_sensor_states()
{
   int i;
   int setting;
   char temp_buf[64];

   for (i = 0; sensors[i].formula; i++)
   {
      sprintf(temp_buf, "sensor%i", i);
      setting = get_config_int("sensors", temp_buf, 0x12345678);
      if (setting != 0x12345678)
         sensors[i].enabled = setting;
   }
}


void save_sensor_states()
{
   int i;
   char temp_buf[64];

   for (i = 0; sensors[i].formula; i++)
   {
      sprintf(temp_buf, "sensor%i", i);
      set_config_int("sensors", temp_buf, sensors[i].enabled);
   }
}
/* ---- TO HERE ---- */

void fill_sensors(int page_number)
{
   int i;
   int index = 0;
   int sensor_index;
   const char *def_str;

   for (i = 0; sensor_dialog[i].proc; i++)
   {
      if (sensor_dialog[i].proc == sensor_proc)
      {
         sensor_index = index + page_number * SENSORS_PER_PAGE;
         if (sensors[sensor_index].formula)
         {
            if (sensors[sensor_index].enabled)
               def_str = "N/A";
            else
               def_str = "not monitoring";
            if (!sensors[sensor_index].supported) def_str = "NOT SUPPORTED!";
            strcpy(sensors[sensor_index].screen_buf, def_str);
            sensor_dialog[i].dp3 = &sensors[sensor_index];
            index++;
         }
         else
         {
            sensor_dialog[i].dp3 = NULL;
            sensor_dialog[i].flags |= D_DIRTY;
         }
      }
   }

   sensors_on_page = index;
}


int page_number_proc(int msg, DIALOG *d, int c)
{
   switch (msg)
   {
      case MSG_START:
         if ((d->dp = calloc(64, sizeof(char))) == NULL)
            fatal_error("Could not allocate enough memory for page number control"); // do not translate
      case MSG_UPDATE:
         sprintf(d->dp, "%i of %i", current_page + 1, ((num_of_sensors % SENSORS_PER_PAGE) ? num_of_sensors/SENSORS_PER_PAGE + 1 : num_of_sensors/SENSORS_PER_PAGE));
         d->flags |= D_DIRTY;
         break;

      case MSG_DRAW:
         rectfill(screen, d->x-d->w, d->y, d->x+d->w-1, d->y+d->h-1, d->bg);  // clear the element
         break;

      case MSG_END:
         free(d->dp);
         d->dp = NULL;
         break;
   }

   return st_ctext_proc(msg, d, c);
}


int page_flipper_proc(int msg, DIALOG *d, int c)
{
   int ret;
   int last_page = (num_of_sensors % SENSORS_PER_PAGE) ? num_of_sensors/SENSORS_PER_PAGE : num_of_sensors/SENSORS_PER_PAGE - 1;

   switch (msg)
   {
      case MSG_START:
      case MSG_UPDATE:
         if (d->d1 == -1)
         {
            if (current_page <= 0)
               d->flags |= D_DISABLED;
            else
               d->flags &= ~D_DISABLED;
         }
         else if (d->d1 == 1)
         {
            if (current_page >= last_page)
               d->flags |= D_DISABLED;
            else
               d->flags &= ~D_DISABLED;
         }
         d->flags |= D_DIRTY;
         break;
   }

   ret = d_button_proc(msg, d, c);

   if (ret == D_CLOSE)
   {
      if (d->d1 == -1)
         current_page--;
      else if (d->d1 == 1)
         current_page++;

      if (current_page < 0)
         current_page = 0;
      else if (current_page > last_page)
         current_page = last_page;

      fill_sensors(current_page);
      inst_refresh_rate = -1;
      avg_refresh_rate = -1;
      broadcast_dialog_message(MSG_UPDATE, 0);

      ret = D_REDRAWME;
   }

   return ret;
}


int toggle_proc(int msg, DIALOG *d, int c)
{
   int ret;

   switch (msg)
   {
      case MSG_START:
         d->d2 = 1;
         if ((d->dp = calloc(4, sizeof(char))) == NULL)
         {
            // don't translate!
            sprintf(temp_error_buf, "Could not allocate enough memory for toggle button caption: %i", d->d1);
            fatal_error(temp_error_buf);
         }
         strcpy(d->dp, "ON");
         d->bg = C_GREEN;
         // Fall through

      case MSG_UPDATE:
         if (d->d1 >= sensors_on_page)
         {
            d->flags |= D_DISABLED;
            d->bg = C_LIGHT_GRAY;
            strcpy(d->dp, empty_string);
            if (d->d2 == 0)
               num_of_disabled_sensors--;
         }
         else
         {
            if ((d->flags & D_DISABLED) && (d->d2 == 0))
               num_of_disabled_sensors++;
            d->flags &= ~D_DISABLED;
            if (d->d2 == 1)
            {
               d->bg = C_GREEN;
               strcpy(d->dp, "ON");
            }
            else
            {
               d->bg = C_DARK_YELLOW;
               strcpy(d->dp, "OFF");
            }
         }
         d->flags |= D_DIRTY;
         break;

      case MSG_TOGGLE:
         if (((c == d->d1) || ((c == -2) && (d->d2 == 1)) || ((c == -1) && (d->d2 == 0))) && !(d->flags & D_DISABLED))
         {
            if (d->d2 == 0)
            {
               d->d2 = 1;
               strcpy(d->dp, "ON");
               d->bg = C_GREEN;
            }
            else
            {
               d->d2 = 0;
               strcpy(d->dp, "OFF");
               d->bg = C_DARK_YELLOW;
            }

            return D_REDRAWME;
         }
         break;

      case MSG_END:
         free(d->dp);
         d->dp = NULL;
         break;
   }

   ret = d_button_proc(msg, d, c);

   if (ret == D_CLOSE)
   {
      broadcast_dialog_message(MSG_TOGGLE, d->d1);
      return D_O_K;
   }

   return ret;
}


int toggle_all_proc(int msg, DIALOG *d, int c)
{
   int ret;

   switch (msg)
   {
      case MSG_START:
         d->d2 = 1;
         if ((d->dp = calloc(64, sizeof(char))) == NULL)
            fatal_error("Could not allocate enough memory for toggle_all button caption"); // do not translate
         strcpy(d->dp, "&All OFF");
         // Fall through

      case MSG_UPDATE:
         if (sensors_on_page == 0)
         {
            d->flags |= D_DISABLED;
            d->bg = C_LIGHT_GRAY;
         }
         else
         {
            d->flags &= ~D_DISABLED;
            if (d->bg == C_LIGHT_GRAY)
               d->bg = C_DARK_YELLOW;
            object_message(d, MSG_TOGGLE, 0);
         }
         d->flags |= D_DIRTY;
         break;

      case MSG_TOGGLE:
         if ((c == -1) || ((c >= 0) && (num_of_disabled_sensors <= 0)))
         {
            d->d2 = 1;
            strcpy(d->dp, "&All OFF");
            d->bg = C_DARK_YELLOW;
         }
         else if ((c == -2) || ((c >= 0) && (num_of_disabled_sensors >= sensors_on_page)))
         {
            d->d2 = 0;
            strcpy(d->dp, "&All ON");
            d->bg = C_GREEN;
            inst_refresh_rate = -1;
            avg_refresh_rate = -1;
            broadcast_dialog_message(MSG_REFRESH, 0);
         }
         d->flags |= D_DIRTY;
         break;

      case MSG_END:
         free(d->dp);
         d->dp = NULL;
         break;
   }

   ret = d_button_proc(msg, d, c);

   if (ret == D_CLOSE)
   {
      if (d->d2 == 0)
         broadcast_dialog_message(MSG_TOGGLE, -1);
      else
         broadcast_dialog_message(MSG_TOGGLE, -2);

      ret = D_REDRAWME;
   }

   return ret;
}


int status_proc(int msg, DIALOG *d, int c)
{
   switch (msg)
   {
      case MSG_START:
         if((d->dp = calloc(48, sizeof(char))) == NULL)
            fatal_error("Could not allocate enough memory for status control (sensors)");
         d->d1 = device_connected;
         d->d2 = -1;
         object_message(d, MSG_IDLE, 0);
         break;

      case MSG_END:
         free(d->dp);
         d->dp = NULL;
         break;

      case MSG_DRAW:
      {
         int circle_color;

         if ((comport.status == READY) && device_connected)
            circle_color = C_GREEN;
         else if ((comport.status == NOT_OPEN) || (comport.status == USER_IGNORED))
            circle_color = C_RED;
         else
            circle_color = C_DARK_YELLOW;

         circlefill(screen, d->x-d->h/2, d->y+d->h/2-3, d->h/2-2, circle_color);
         circle(screen, d->x-d->h/2, d->y+d->h/2-3, d->h/2-2, C_BLACK);

         rectfill(screen, d->x, d->y, d->x+d->w-1, d->y+d->h-1, d->bg);  // clear the text area
         break;
      }
      case MSG_IDLE:
         if ((d->d1 != device_connected) || (d->d2 != comport.status))
         {
            d->d1 = device_connected;
            d->d2 = comport.status;

            if (comport.status == READY)
            {
               if (device_connected)
                  sprintf(d->dp, " COM%i ready (device connected)", comport.number + 1);
               else
                  sprintf(d->dp, " COM%i ready (device not found)", comport.number + 1);
            }
            else
               sprintf(d->dp, " COM%i could not be opened", comport.number + 1);

            return D_REDRAWME;
         }
   }

   return d_text_proc(msg, d, c);
}


#define SENSOR_LABEL_MARGIN   245
#define SENSOR_VALUE_INDENT   8

int sensor_proc(int msg, DIALOG *d, int c)
{
   static int current_sensor = 0; // sensor we're working on
   static int new_page = FALSE;
   static int receiving_response = FALSE; // flag: receiving or sending
   static int num_of_sensors_timed_out = 0;
   static int ignore_device_not_connected = FALSE;
   static int retry_attempts = NUM_OF_RETRIES;
   static int active_sensor_found = FALSE;
   static char vehicle_response[1024];
   char buf[256];
   char cmd[8];
   int response_status = EMPTY; //status of the response: EMPTY, DATA, PROMPT
   int response_type;
   int ret = 0;
   SENSOR *sensor = (SENSOR *)d->dp3; // create a pointer to SENSOR structure in dp3 (vm)

   if ((msg == MSG_IDLE) && reset_hardware && comport.status == READY) // if user hit "Reset Chip" button, and we're doing nothing
   {
      reset_hardware = FALSE;
      receiving_response = FALSE;
      device_connected = FALSE;

      reset_chip();
      return D_O_K;
   }

   if (!sensor && (msg != MSG_DRAW))  // don't wanna handle any messages for an empty sensor line except MSG_DRAW
      return D_O_K;

   switch (msg)
   {
      case MSG_START:
         if (d->d1 == 0)
         {
            stop_serial_timer();
            receiving_response = FALSE;
            device_connected = FALSE;
            ignore_device_not_connected = FALSE;
            num_of_disabled_sensors = 0;
         }
         d->d2 = 0;
         d->flags &= ~D_DISABLED;
         // fall through

      case MSG_UPDATE:
         if (d->d1 == 0)  // if it's the first sensors in sensor_dialog
         {
            current_sensor = 0;
            num_of_sensors_timed_out = 0;
            active_sensor_found = FALSE;
            new_page = TRUE;
            // page was flipped, reset refresh rate variables
            inst_refresh_rate = 0;
            avg_refresh_rate = 0;
            refresh_time = 0;
         }
         if ((sensor->enabled && d->flags & D_DISABLED) || (!sensor->enabled && !(d->flags & D_DISABLED)))
            d->d2 = 1;
         d->flags |= D_DIRTY;
         break;

      case MSG_END:
         if (d->d1 == sensors_on_page - 1)
            stop_serial_timer();
         d->dp3 = NULL;
         break;

      case MSG_TOGGLE:
         if((d->d1 == c) || ((c == -2) && !(d->flags & D_DISABLED)) || ((c == -1) && (d->flags & D_DISABLED)))
         {
            if (d->flags & D_DISABLED)
            {
               d->flags &= ~D_DISABLED;
               sensor->enabled = TRUE;
               num_of_disabled_sensors--;
               strcpy(sensor->screen_buf, "N/A");
            }
            else
            {
               d->flags |= D_DISABLED;
               sensor->enabled = FALSE;
               num_of_disabled_sensors++;
               strcpy(sensor->screen_buf, "not monitoring");
               if (d->d1 == current_sensor)
               {
                  receiving_response = FALSE;
                  stop_serial_timer();
               }
               if (num_of_disabled_sensors == sensors_on_page) // if all of the sensors are disabled
                  num_of_sensors_timed_out = 0; // reset timeout counter
            }
            if (!sensor->supported) strcpy(sensor->screen_buf, "NOT SUPPORTED!");

            return D_REDRAWME;
         }
         break;

      case MSG_DRAW:
         if (d->d2)
         {
            d->d2 = 0;
            broadcast_dialog_message(MSG_TOGGLE, d->d1);
         }
         rectfill(screen, d->x, d->y, d->x+d->w-1, d->y+d->h-1, d->bg);  // clear the element
         if (sensor)
         {
            //gui_textout_ex(screen, sensor->label, d->x + SENSOR_LABEL_MARGIN - gui_strlen(sensor->label), d->y, d->fg, d->bg, FALSE);
            // dle changed label length to fixed 28 char length, so sensor output is left justified
            gui_textout_ex(screen, sensor->label, d->x + SENSOR_LABEL_MARGIN - gui_strlen("1234567890123456789012345678"), d->y, d->fg, d->bg, FALSE);
            gui_textout_ex(screen, sensor->screen_buf, d->x + SENSOR_LABEL_MARGIN + SENSOR_VALUE_INDENT, d->y, ((d->flags & D_DISABLED) ? gui_mg_color : d->fg), d->bg, FALSE);
         }
         return D_O_K;

      case MSG_IDLE:
         if (d->d1 == current_sensor)
         {
            if (comport.status == READY)
            {
               if (d->flags & D_DISABLED)
                  calculate_refresh_rate(SENSOR_OFF); // calculate instantaneous/average refresh rates

               if (!receiving_response)
               {
                  if (d->flags & D_DISABLED) // if the sensor is disabled
                  {
                     current_sensor = (current_sensor == sensors_on_page - 1) ? 0 : current_sensor + 1; // go to the next sensor
                     return D_O_K;
                  }

                  sprintf(cmd, "01%s", sensor->pid);
                  send_command(cmd); // send command for that particular sensor
                  new_page = FALSE;
                  receiving_response = TRUE; // now we're waiting for response
                  start_serial_timer(OBD_REQUEST_TIMEOUT); // start the timer
                  vehicle_response[0] = '\0'; // clear string
               }
               else  // if we are receiving response
               {
                  response_status = read_comport(buf);  // read comport

                  if (d->flags & D_DISABLED) // if the sensor is disabled
                  {
                     if (response_status == PROMPT)
                        receiving_response = FALSE;
                     current_sensor = (current_sensor == sensors_on_page - 1) ? 0 : current_sensor + 1; // go to the next sensor
                     return D_O_K;
                  }

                  if (response_status == DATA) // if data detected in com port buffer
                  {
                     strcat(vehicle_response, buf); // append contents of buf to vehicle_response
                  }
                  else if (response_status == PROMPT) // if '>' detected
                  {
                     device_connected = TRUE;
                     num_of_sensors_timed_out = 0;
                     receiving_response = FALSE; // we're not waiting for response any more
                     stop_serial_timer();

                     if (new_page)
                        break;

                     strcat(vehicle_response, buf); // append contents of buf to vehicle_response
                     sprintf(cmd, "01%s", sensor->pid);
                     response_type = process_response(cmd, vehicle_response);

                     if (response_type == HEX_DATA)  // HEX_DATA received
                     {
                        sprintf(cmd, "41%s", sensor->pid);
                        if (find_valid_response(buf, vehicle_response, cmd, NULL))
                        {
                           active_sensor_found = TRUE;
                           calculate_refresh_rate(SENSOR_ACTIVE); // calculate instantaneous/average refresh rates
                           buf[4 + sensor->bytes * 2] = 0;  // solves problem where response is padded with zeroes (i.e., '41 05 7C 00 00 00')
                           sensor->formula((int)strtol(buf + 4, NULL, 16), buf); //plug the value into formula
                           iindex = current_sensor + (current_page * SENSORS_PER_PAGE);
                           strcpy(sensor->screen_buf, buf);  // and copy result in screen buffer
                           // dle adding to log data from sensors when collecting data, and in data collection state
                           // buf1 is value of parameter minus text
                           if (iGraphState == 1) write_datalog(buf1);
                           /* if current_sensor is the last sensor, set it to 0; otherwise current_sensor++ */
                           current_sensor = (current_sensor == sensors_on_page - 1) ? 0 : current_sensor + 1;
                           retry_attempts = NUM_OF_RETRIES;

                           return D_REDRAWME;
                        }
                        else
                           response_type = ERR_NO_DATA;
                     }

                     strcpy(sensor->screen_buf, "N/A");

                     if (active_sensor_found)
                        calculate_refresh_rate(SENSOR_NA); // calculate instantaneous/average refresh rates

                     if (response_type == ERR_NO_DATA) // if we received "NO DATA", "N/A" will be printed
                     {
                        current_sensor = (current_sensor == sensors_on_page - 1) ? 0 : current_sensor + 1; // next time poll next sensor
                        retry_attempts = NUM_OF_RETRIES;
                     }
                     else if (response_type == BUS_ERROR || response_type == UNABLE_TO_CONNECT || response_type == BUS_INIT_ERROR)
                     {
                        display_error_message(response_type, FALSE);
                        retry_attempts = NUM_OF_RETRIES;
                     }
                     else // for other errors, try to re-send the request, do nothing if successful and alert user if failed
                     {
                        if (retry_attempts > 0)
                        {
                           retry_attempts--;
                           return D_O_K;
                        }
                        else
                        {
                           display_error_message(response_type, FALSE);
                           retry_attempts = NUM_OF_RETRIES; // reset the number of retry attempts
                        }
                     }

                     return D_REDRAWME;
                  }
                  else if (serial_time_out) // if timeout occured,
                  {
                     receiving_response = FALSE; // we're not waiting for a response any more
                     strcpy(sensor->screen_buf, "N/A");

                     if (num_of_sensors_timed_out >= SENSORS_TO_TIME_OUT)
                     {
                        num_of_sensors_timed_out = 0;
                        device_connected = FALSE;
                        if  (!ignore_device_not_connected)
                        {
                           ret = alert3("Device is not responding.", "Please check that it is connected", "and the port settings are correct", "&OK", "&Configure Port", "&Ignore", 'o', 'c', 'i');
                           if (ret == 2)
                              display_options();   // let the user choose correct settings
                           else if (ret == 3)
                              ignore_device_not_connected = TRUE;
                        }
                     }
                     else
                        num_of_sensors_timed_out++;

                     while (comport.status == NOT_OPEN)
                     {
                        if (alert("Port is not ready.", "Please check that you specified the correct port", "and that no other application is using it", "&Configure Port", "&Ignore", 'c', 'i') == 1)
                           display_options(); // let the user choose correct settings
                        else
                           comport.status = USER_IGNORED;
                     }

                     stop_serial_timer();

                     /* if current_sensor is the last sensor, set it to 0; otherwise current_sensor++ */
                     current_sensor = (current_sensor == sensors_on_page - 1) ? 0 : current_sensor + 1;

                     return D_REDRAWME;
                  }
               }
            }
         }
         break;
   } // end of switch (msg)

   return D_O_K;
}


void engine_rpm_formula(int data, char *buf)
{
//   if (system_of_measurements == METRIC)
//      sprintf(buf, "%i r/min", (int)((float)data/4));
//   else   // if the system is IMPERIAL
   sprintf(buf, "%i RPM", (int)((float)data/4));
   sprintf(buf1, "  %i", (int)((float)data/4));
   sprintf(title1, "RPM");
   sprintf(title2, " ");
}


void engine_load_formula(int data, char *buf)
{
   sprintf(buf, "%.1f%%", (float)data*100/255);
   sprintf(buf1, "  %.1f", (float)data*100/255);
   sprintf(title1, "Percent");
   sprintf(title2, " ");
}


void coolant_temp_formula(int data, char *buf)
{
   if (system_of_measurements == METRIC) {
      sprintf(buf, "%i%c C", data-40, 0xB0);
      sprintf(buf1, "  %i", data-40);
      sprintf(title1, "Degrees C");
      sprintf(title2, " ");
   }
   else {  // if the system is IMPERIAL
      sprintf(buf, "%i%c F", (int)(((float)data-40)*9/5 + 32), 0xB0);
      sprintf(buf1, "  %i", (int)(((float)data-40)*9/5 + 32));
      sprintf(title1, "Degrees F");
      sprintf(title2, " ");
   }
}

// dle this needs work
void fuel_system_status_formula(int data, char *buf)
{
   if (data == 0)
      sprintf(buf, "unused");
   else if (data == 0x01)
      sprintf(buf, "open loop");
   else if (data == 0x02)
      sprintf(buf, "closed loop");
   else if (data == 0x04)
      sprintf(buf, "open loop (driving conditions)");
   else if (data == 0x08)
      sprintf(buf, "open loop (system fault)");
   else if (data == 0x10)
      sprintf(buf, "closed loop, O2 sensor fault");
   else
      sprintf(buf, "unknown: 0x%02X", data);
   sprintf(buf1, "N/A");
   sprintf(title1, " ");
   sprintf(title2, " ");
}

/*
void fuel_system1_status_formula(int data, char *buf)
{
   fuel_system_status_formula((data >> 8) & 0xFF, buf);  // Fuel System 1 status: Data A
   fuel_system_status_formula((data >> 8) & 0xFF, buf1);  // Fuel System 1 status: Data A
}
*/

void fuel_system2_status_formula(int data, char *buf)
{
   fuel_system_status_formula(data & 0xFF, buf);  // Fuel System 2 status: Data B
   fuel_system_status_formula(data & 0xFF, buf1);  // Fuel System 2 status: Data B
}


void vehicle_speed_formula(int data, char *buf)
{
   if (system_of_measurements == METRIC) {
      sprintf(buf, "%i Km/h", data);
      sprintf(buf1, "  %i", data);
      sprintf(title1, "Km/h");
      sprintf(title2, " ");
   }
   else {  // if the system is IMPERIAL
      sprintf(buf, "%i MPH", (int)(data/1.609));
      sprintf(buf1, "  %i", (int)(data/1.609));
      sprintf(title1, "MPH");
      sprintf(title2, " ");
   }
}


void intake_pressure_formula(int data, char *buf)
{
   if (system_of_measurements == METRIC) {
      sprintf(buf, "%i kPa", data);
      sprintf(buf1, "  %i", data);
      sprintf(title1, "kPa");
      sprintf(title2, " ");
   }
   else {
      sprintf(buf, "%.1f inHg", data/3.386389);
      sprintf(buf1, "  %.1f", data/3.386389);
      sprintf(title1, "inHg");
      sprintf(title2, " ");
   }
}


void timing_advance_formula(int data, char *buf)
{
   sprintf(buf, "%.1f%c", ((float)data-128)/2, 0xB0);
   sprintf(buf1, "  %.1f", ((float)data-128)/2);
   sprintf(title1, "Degrees");
   sprintf(title2, " ");
}


void intake_air_temp_formula(int data, char *buf)
{
   if (system_of_measurements == METRIC) {
      sprintf(buf, "%i%c C", data-40, 0xB0);
      sprintf(buf1, "  %i", data-40);
      sprintf(title1, "Degrees C");
      sprintf(title2, " ");
   }
   else {  // if the system is IMPERIAL
      sprintf(buf, "%i%c F", (int)(((float)data-40)*9/5 + 32), 0xB0);
      sprintf(buf1, "  %i", (int)(((float)data-40)*9/5 + 32));
      sprintf(title1, "Degrees F");
      sprintf(title2, " ");
   }
}


void air_flow_rate_formula(int data, char *buf)
{
   if (system_of_measurements == METRIC) {
      sprintf(buf, "%.2f g/s", data*0.01);
      sprintf(buf1, "  %.2f", data*0.01);
      sprintf(title1, "g/s");
      sprintf(title2, " ");
   }
   else {
      sprintf(buf, "%.1f lb/min", data*0.0013227736);
      sprintf(buf1, "  %.1f", data*0.0013227736);
      sprintf(title1, "lb/min");
      sprintf(title2, " ");
   }
}


void throttle_position_formula(int data, char *buf)
{
   sprintf(buf, "%.1f%%", (float)data*100/255);
   sprintf(buf1, "  %.1f", (float)data*100/255);
   sprintf(title1, "Percent");
   sprintf(title2, " ");
}


// **** New formulae added 3/11/2003: ****

// Fuel Pressure (gauge): PID 0A
void fuel_pressure_formula(int data, char *buf)
{
   if (system_of_measurements == METRIC) {
      sprintf(buf, "%i kPa", data*3);
      sprintf(buf1, "  %i", data*3);
      sprintf(title1, "kPa");
      sprintf(title2, " ");
   }
   else {  // if the system is IMPERIAL
      sprintf(buf, "%.1f PSI", data*3*0.145037738);
      sprintf(buf1, "  %.1f", data*3*0.145037738);
      sprintf(title1, "PSI");
      sprintf(title2, " ");
   }
}


// Fuel Trim statuses: PID 06-09
void short_term_fuel_trim_formula(int data, char *buf)
{
   if (data > 0xFF)  // we're only showing bank 1 and 2 FT
      data >>= 8;

   sprintf(buf, (data == 128) ? "0.0%%" : "%+.1f%%", ((float)data - 128)*100/128);
   sprintf(buf1, (data == 128) ? "   0.0" : "  %+.1f", ((float)data - 128)*100/128);
   sprintf(title1, "Percent");
   sprintf(title2, " ");
}


void long_term_fuel_trim_formula(int data, char *buf)
{
   if (data > 0xFF)  // we're only showing bank 1 and 2 FT
      data >>= 8;

   sprintf(buf, (data == 128) ? "0.0%%" : "%+.1f%%", ((float)data - 128)*100/128);
   sprintf(buf1, (data == 128) ? "  0.0" : " %+.1f", ((float)data - 128)*100/128);
   sprintf(title1, "Percent");
   sprintf(title2, " ");
}


// Commanded secondary air status: PID 12
void secondary_air_status_formula(int data, char *buf)
{
   data = data & 0x0700; // mask bits 0, 1, and 2

   if (data == 0x0100)
      sprintf(buf, "upstream of 1st cat. conv.");
   else if (data == 0x0200)
      sprintf(buf, "downstream of 1st cat. conv.");
   else if (data == 0x0400)
      sprintf(buf, "atmosphere / off");
   else
      sprintf(buf, "Not supported");
   sprintf(buf1, "N/A");
   sprintf(title1, " ");
   sprintf(title2, " ");
}

// Oxygen sensor voltages & short term fuel trims: PID 14-1B
// Format is bankX_sensor

void o2_sensor_formula(int data, char *buf)
{
   if ((data & 0xFF) == 0xFF) { // if the sensor is not used in fuel trim calculation,
      sprintf(buf, "%.3f Volts", (data >> 8)*0.005);
      sprintf(buf1, "  %.3f", (data >> 8)*0.005);
      sprintf(title1, "Volts");
      sprintf(title2, " ");
   }
   else {
      sprintf(buf, ((data & 0xFF) == 128) ? "%.3f V @ 0.0%% S.T. Fuel Trim" : "%.3f V @ %+.1f%% S.T. Fuel Trim", (data >> 8)*0.005, ((float)(data & 0xFF) - 128)*100/128);
      // note extra space below for alignment on output
      sprintf(buf1, ((data & 0xFF) == 128) ? "  %.3f   0.0" : "  %.3f  %+.1f", (data >> 8)*0.005, ((float)(data & 0xFF) - 128)*100/128);
      sprintf(title1, "Volts");
      sprintf(title2, "Percent S.T. Fuel Trim");
   }
}


//Power Take-Off Status: PID 1E
void pto_status_formula(int data, char *buf)
{
   if ((data & 0x01) == 0x01)
      sprintf(buf, "active");
   else
      sprintf(buf, "not active");
   sprintf(buf1, "N/A");
   sprintf(title1, " ");
   sprintf(title2, " ");
}

// OBD requirement to which vehicle is designed: PID 1C
void obd_requirements_formula(int data, char *buf)
{
   switch (data)
   {
      case 0x01:
         sprintf(buf, "OBD-II (California ARB)");
         break;
      case 0x02:
         sprintf(buf, "OBD (Federal EPA)");
         break;
      case 0x03:
         sprintf(buf, "OBD and OBD-II");
         break;
      case 0x04:
         sprintf(buf, "OBD-I");
         break;
      case 0x05:
         sprintf(buf, "Not OBD compliant");
         break;
      case 0x06:
         sprintf(buf, "EOBD (Europe)");
         break;
      case 0x07:
         sprintf(buf, "EOBD and OBD-II");
         break;
      case 0x08:
         sprintf(buf, "EOBD and OBD");
         break;
      case 0x09:
         sprintf(buf, "EOBD, OBD and OBD-II");
         break;
      case 0x0A:
         sprintf(buf, "JOBD (Japan)");
         break;
      case 0x0B:
         sprintf(buf, "JOBD and OBD-II");
         break;
      case 0x0C:
         sprintf(buf, "JOBD and EOBD");
         break;
      case 0x0D:
         sprintf(buf, "JOBD, EOBD, and OBD-II");
         break;
      default:
         sprintf(buf, "Unknown: 0x%02X", data);
   }
   sprintf(buf1, "N/A");
   sprintf(title1, " ");
   sprintf(title2, " ");
}

/* Sensors added 1/2/2003: */

void engine_run_time_formula(int data, char *buf)
{
   int sec, min, hrs;

   hrs = data / 3600;  // get hours
   min = (data % 3600) / 60;  // get minutes
   sec = data % 60;  // get seconds

   sprintf(buf, "%02i:%02i:%02i", hrs, min, sec);
   sprintf(buf1, "N/A");
   sprintf(title1, " ");
   sprintf(title2, " ");
}


void mil_distance_formula(int data, char *buf)
{
   if (system_of_measurements == METRIC) {
      sprintf(buf, "%i Km", data);
      sprintf(buf1, "  %i", data);
      sprintf(title1, "Km");
      sprintf(title2, " ");
   }
   else {  // if the system is IMPERIAL
      sprintf(buf, "%i Miles", (int)(data/1.609));
      sprintf(buf1, "  %i", (int)(data/1.609));
      sprintf(title1, "Miles");
      sprintf(title2, " ");
   }
}


void frp_relative_formula(int data, char *buf)
{
   float kpa, psi;

   kpa = data*0.079;
   psi = kpa*0.145037738;

   if (system_of_measurements == METRIC) {
      sprintf(buf, "%.3f kPa", kpa);
      sprintf(buf1, "  %.3f", kpa);
      sprintf(title1, "kPa");
      sprintf(title2, " ");
   }
   else {  // if the system is IMPERIAL
      sprintf(buf, "%.1f PSI", psi);
      sprintf(buf1, "  %.1f", psi);
      sprintf(title1, "PSI");
      sprintf(title2, " ");
   }
}


void frp_widerange_formula(int data, char *buf)
{
   int kpa;
   float psi;

   kpa = data*10;
   psi = kpa*0.145037738;

   if (system_of_measurements == METRIC) {
      sprintf(buf, "%i kPa", kpa);
      sprintf(buf1, "  %i", kpa);
      sprintf(title1, "kPa");
      sprintf(title2, " ");
   }
   else {
      sprintf(buf, "%.1f PSI", psi);
      sprintf(buf1, "  %.1f", psi);
      sprintf(title1, "PSI");
      sprintf(title2, " ");
   }
}


void o2_sensor_wrv_formula(int data, char *buf)
{
   float eq_ratio, o2_voltage; // equivalence ratio and sensor voltage

   eq_ratio = (float)(data >> 16)*0.0000305;  // data bytes A,B
   o2_voltage = (float)(data & 0xFFFF)*0.000122; // data bytes C,D

   sprintf(buf, "%.3f V, Eq. ratio: %.3f", o2_voltage, eq_ratio);
   sprintf(buf1, "  %.3f  %.3f", o2_voltage, eq_ratio);
   sprintf(title1, "Volts");
   sprintf(title2, "Equivalence Ratio");
}


//Commanded EGR status: PID 2C
void commanded_egr_formula(int data, char *buf)
{
   sprintf(buf, "%.1f%%", (float)data*100/255);
   sprintf(buf1, "  %.1f", (float)data*100/255);
   sprintf(title1, "Percent");
   sprintf(title2, " ");
}

//EGR error: PID 2D
void egr_error_formula(int data, char *buf)
{
   sprintf(buf, (data == 128) ? "0.0%%" : "%+.1f%%", (float)(data-128)/255*100);
   sprintf(buf1, (data == 128) ? "   0.0" : "  %+.1f", (float)(data-128)/255*100);
   sprintf(title1, "Percent");
   sprintf(title2, "Percent");
}


void evap_pct_formula(int data, char *buf)
{
   sprintf(buf, "%.1f%%", (float)data/255*100);
   sprintf(buf1, "  %.1f", (float)data/255*100);
   sprintf(title1, "Percent");
   sprintf(title2, " ");
}


void fuel_level_formula(int data, char *buf)
{
   sprintf(buf, "%.1f%%", (float)data/255*100);
   sprintf(buf1, "  %.1f", (float)data/255*100);
   sprintf(title1, "Percent");
   sprintf(title2, " ");
}


void warm_ups_formula(int data, char *buf)
{
   sprintf(buf, "%i", data);
   sprintf(buf1, "  %i", data);
   sprintf(title1, " ");
   sprintf(title2, " ");
}


void clr_distance_formula(int data, char *buf)
{
   if (system_of_measurements == METRIC) {
      sprintf(buf, "%i Km", data);
      sprintf(buf1, "  %i", data);
      sprintf(title1, "Km");
      sprintf(title2, " ");
   }
   else {
      sprintf(buf, "%i Miles", (int)(data/1.609));
      sprintf(buf1, "  %i", (int)(data/1.609));
      sprintf(title1, "Miles");
      sprintf(title2, " ");
   }
}


void evap_vp_formula(int data, char *buf)
{
   if (system_of_measurements == METRIC) {
      sprintf(buf, "%.2f Pa", data*0.25);
      sprintf(buf1, "  %.2f", data*0.25);
      sprintf(title1, "Pa");
      sprintf(title2, " ");
   }
   else {
      sprintf(buf, "%.3f inches H2O", data*0.25/249.088908);
      sprintf(buf1, "  %.3f", data*0.25/249.088908);
      sprintf(title1, "inches H20");
      sprintf(title2, " ");
   }
}


void baro_pressure_formula(int data, char *buf)
{
   if (system_of_measurements == METRIC) {
      sprintf(buf, "%i kPa", data);
      sprintf(buf1, "  %i", data);
      sprintf(title1, "kPa");
      sprintf(title2, " ");
   }
   else {
      sprintf(buf, "%.1f inches Hg", data*0.2953);
      sprintf(buf1, "  %.1f", data*0.2953);
      sprintf(title1, "inches Hg");
      sprintf(title2, " ");
   }
}


void o2_sensor_wrc_formula(int data, char *buf)
{
   float eq_ratio, o2_ma; // equivalence ratio and sensor current

   eq_ratio = (float)(data >> 16)*0.0000305;  // data bytes A,B
   o2_ma = ((float)(data & 0xFFFF) - 0x8000)*0.00390625; // data bytes C,D

   sprintf(buf, "%.3f mA, Eq. ratio: %.3f", o2_ma, eq_ratio);
   sprintf(buf1, "  %.3f  %.3f", o2_ma, eq_ratio);
   sprintf(title1, "mA");
   sprintf(title2, "Equivalence Ratio");
}


void cat_temp_formula(int data, char *buf)
{
   float c, f;

   c = data*0.1 - 40; // degrees Celcius
   f = c*9/5 + 32;  // degrees Fahrenheit

   if (system_of_measurements == METRIC) {
      sprintf(buf, "%.1f%c C", c, 0xB0);
      sprintf(buf1, "  %.1f", c);
      sprintf(title1, "Degrees C");
      sprintf(title2, " ");
   }
   else {
      sprintf(buf, "%.1f%c F", f, 0xB0);
      sprintf(buf1, "  %.1f", f);
      sprintf(title1, "Degrees F");
      sprintf(title2, " ");
   }
}


void ecu_voltage_formula(int data, char *buf)
{
   sprintf(buf, "%.3f Volts", data*0.001);
   sprintf(buf1, "  %.3f", data*0.001);
   sprintf(title1, "Volts");
   sprintf(title2, " ");
}


void abs_load_formula(int data, char *buf)
{
   sprintf(buf, "%.1f%%", (float)data*100/255);
   sprintf(buf1, "  %.1f", (float)data*100/255);
   sprintf(title1, "Percent");
   sprintf(title2, " ");
}


void eq_ratio_formula(int data, char *buf)
{
   sprintf(buf, "%.3f", data*0.0000305);
   sprintf(buf1, "  %.3f", data*0.0000305);
   sprintf(title1, " ");
   sprintf(title2, " ");
}


void relative_tp_formula(int data, char *buf)
{
   sprintf(buf, "%.1f%%", (float)data*100/255);
   sprintf(buf1, "  %.1f", (float)data*100/255);
   sprintf(title1, "Percent");
   sprintf(title2, " ");
}


void amb_air_temp_formula(int data, char *buf)
{
   int c, f;

   c = data-40; // degrees Celcius
   f = (float)c*9/5 + 32;  // degrees Fahrenheit

   if (system_of_measurements == METRIC) {
      sprintf(buf, "%i%c C", c, 0xB0);
      sprintf(buf1, "  %i", c);
      sprintf(title1, "Degrees C");
      sprintf(title2, " ");
   }
   else {
      sprintf(buf, "%i%c F", f, 0xB0);
      sprintf(buf1, "  %i", f);
      sprintf(title1, "Degrees F");
      sprintf(title2, " ");
   }
}


void abs_tp_formula(int data, char *buf)
{
   sprintf(buf, "%.1f%%", (float)data*100/255);
   sprintf(buf1, "  %.1f", (float)data*100/255);
   sprintf(title1, "Percent");
   sprintf(title2, " ");
}


void tac_pct_formula(int data, char *buf)
{
   sprintf(buf, "%.1f%%", (float)data*100/255);
   sprintf(buf1, "  %.1f", (float)data*100/255);
   sprintf(title1, "Percent");
   sprintf(title2, " ");
}


void mil_time_formula(int data, char *buf)
{
   sprintf(buf, "%i hrs %i min", data/60, data%60);
   sprintf(buf1, "N/A");
   sprintf(title1, " ");
   sprintf(title2, " ");
}


void clr_time_formula(int data, char *buf)
{
   sprintf(buf, "%i hrs %i min", data/60, data%60);
   sprintf(buf1, "N/A");
   sprintf(title1, "RPM");
   sprintf(title2, " ");
}


// dle needs work
void oxygen_sensors_pres_formula(int data, char *buf)
{
   sprintf(buf, "N/A");
   sprintf(buf1, "N/A");
   sprintf(title1, " ");
   sprintf(title2, " ");
}

// dle needs work
void monitor_status_DC_formula(int data, char *buf)
{
   sprintf(buf, "N/A");
   sprintf(buf1, "N/A");
   sprintf(title1, " ");
   sprintf(title2, " ");
}




int genuine_proc(int msg, DIALOG *d, int c)
{
   if (msg == MSG_IDLE)
      if (is_not_genuine_scan_tool == TRUE)
         return D_CLOSE;

   return D_O_K;
}
